/* See https://github.com/geopandas/geopandas/pull/1299 */
/* Copied from sphinx' basic.css to ensure the sphinx >2.0 docstrings are
rendered somewhat properly (xref https://github.com/numpy/numpydoc/issues/215) */

.classifier {
    font-style: oblique;
}

.classifier:before {
    font-style: normal;
    margin: 0.5em;
    content: ":";
}
