.. This (-*- rst -*-) format file contains commonly used link targets
   and name substitutions.  It may be included in many files,
   therefore it should only contain link targets and name
   substitutions.  Try grepping for "^\.\. _" to find plausible
   candidates for this list.

.. NOTE: reST targets are
   __not_case_sensitive__, so only one target definition is needed for
   nipy, NIPY, Nipy, etc...

.. Git stuff
.. _Git: http://git-scm.com/
.. _GitHub: http://github.com
.. _GitHub Help: http://help.github.com
.. _Git for Windows: https://git-for-windows.github.io/
.. _git-osx-installer: https://sourceforge.net/projects/git-osx-installer/
.. _Subversion: http://subversion.tigris.org/
.. _Git cheat sheet: http://github.com/guides/git-cheat-sheet
.. _Pro Git book: http://progit.org/
.. _Git SVN crash course: http://git-scm.com/course/svn.html
.. _learn.github: http://learn.github.com/
.. _network graph visualizer: http://github.com/blog/39-say-hello-to-the-network-graph-visualizer
.. _Git user manual: http://schacon.github.com/git/user-manual.html
.. _Git tutorial: http://schacon.github.com/git/gittutorial.html
.. _Git community book: http://book.git-scm.com/
.. _Git ready: http://www.gitready.com/
.. _Git casts: http://www.gitcasts.com/
.. _Git GUI: https://git-scm.com/downloads/guis
.. _Fernando's Git page: http://www.fperez.org/py4science/git.html
.. _Git magic: http://www-cs-students.stanford.edu/~blynn/gitmagic/index.html
.. _Git concepts: http://www.eecs.harvard.edu/~cduan/technical/git/
.. _git clone: http://schacon.github.com/git/git-clone.html
.. _git checkout: http://schacon.github.com/git/git-checkout.html
.. _git commit: http://schacon.github.com/git/git-commit.html
.. _git push: http://schacon.github.com/git/git-push.html
.. _git pull: http://schacon.github.com/git/git-pull.html
.. _git add: http://schacon.github.com/git/git-add.html
.. _git status: http://schacon.github.com/git/git-status.html
.. _git diff: http://schacon.github.com/git/git-diff.html
.. _git log: http://schacon.github.com/git/git-log.html
.. _git branch: http://schacon.github.com/git/git-branch.html
.. _git remote: http://schacon.github.com/git/git-remote.html
.. _git rebase: http://schacon.github.com/git/git-rebase.html
.. _git config: http://schacon.github.com/git/git-config.html
.. _why the -a flag?: http://www.gitready.com/beginner/2009/01/18/the-staging-area.html
.. _Git staging area: http://www.gitready.com/beginner/2009/01/18/the-staging-area.html
.. _tangled working copy problem: http://tomayko.com/writings/the-thing-about-git
.. _Git management: http://kerneltrap.org/Linux/Git_Management
.. _Linux Git workflow: http://www.mail-archive.com/dri-devel@lists.sourceforge.net/msg39091.html
.. _Git parable: http://tom.preston-werner.com/2009/05/19/the-git-parable.html
.. _Git foundation: http://matthew-brett.github.com/pydagogue/foundation.html
.. _deleting master on GitHub: http://matthew-brett.github.com/pydagogue/gh_delete_master.html
.. _rebase without tears: http://matthew-brett.github.com/pydagogue/rebase_without_tears.html
.. _resolving a merge: http://schacon.github.com/git/user-manual.html#resolving-a-merge
.. _IPython Git workflow: http://mail.scipy.org/pipermail/ipython-dev/2010-October/006746.html

.. other stuff
.. _Python: http://www.python.org
.. _acronym: http://www.numpy.org/devdocs/dev/gitwash/development_workflow.html#writing-the-commit-message

.. |emdash| unicode:: U+02014

.. vim: ft=rst
